@@ -13,6 +13,7 @@
 #include "core/outputlayer.h"
 #include "core/renderbackend.h"
 #include "core/renderloop_p.h"
+#include "wayland-client/viewporter.h"
 #include "wayland_backend.h"
 #include "wayland_display.h"
 #include "wayland_layer.h"
@@ -20,14 +21,16 @@
 #include <KWayland/Client/compositor.h>
 #include <KWayland/Client/pointer.h>
 #include <KWayland/Client/pointerconstraints.h>
+#include <KWayland/Client/seat.h>
 #include <KWayland/Client/surface.h>
 #include <KWayland/Client/xdgdecoration.h>
 
 #include "wayland-fractional-scale-v1-client-protocol.h"
+#include "wayland-keyboard-shortcuts-inhibit-unstable-v1-client-protocol.h"
 #include "wayland-presentation-time-client-protocol.h"
 #include "wayland-single-pixel-buffer-v1-client-protocol.h"
 #include "wayland-tearing-control-v1-client-protocol.h"
-#include "wayland-viewporter-client-protocol.h"
+#include "wayland-xdg-toplevel-icon-v1-client-protocol.h"
 #include "workspace.h"
 
 #include <KLocalizedString>
@@ -46,17 +49,12 @@ using namespace KWayland::Client;
 
 WaylandCursor::WaylandCursor(WaylandBackend *backend)
     : m_surface(backend->display()->compositor()->createSurface())
+    , m_viewport(backend->display()->viewporter()->createViewport(*m_surface))
 {
-    if (auto viewporter = backend->display()->viewporter()) {
-        m_viewport = wp_viewporter_get_viewport(viewporter, *m_surface);
-    }
 }
 
 WaylandCursor::~WaylandCursor()
 {
-    if (m_viewport) {
-        wp_viewport_destroy(m_viewport);
-    }
 }
 
 KWayland::Client::Pointer *WaylandCursor::pointer() const
@@ -100,9 +98,7 @@ void WaylandCursor::sync()
         m_surface->attachBuffer(KWayland::Client::Buffer::Ptr());
         m_surface->commit(KWayland::Client::Surface::CommitFlag::None);
     } else {
-        if (m_viewport) {
-            wp_viewport_set_destination(m_viewport, m_size.width(), m_size.height());
-        }
+        m_viewport->setDestination(m_size);
         m_surface->attachBuffer(m_buffer);
         m_surface->damageBuffer(QRect(0, 0, INT32_MAX, INT32_MAX));
         m_surface->commit(KWayland::Client::Surface::CommitFlag::None);
@@ -123,13 +119,14 @@ const wp_fractional_scale_v1_listener WaylandOutput::s_fractionalScaleListener{
 };
 
 WaylandOutput::WaylandOutput(const QString &name, WaylandBackend *backend)
-    : Output(backend)
+    : BackendOutput()
     , m_renderLoop(std::make_unique<RenderLoop>(this))
     , m_surface(backend->display()->compositor()->createSurface())
     , m_xdgShellSurface(backend->display()->xdgShell()->createSurface(m_surface.get()))
     , m_backend(backend)
     , m_cursor(std::make_unique<WaylandCursor>(backend))
 {
+    m_renderLoop->setMaxPendingFrameCount(2);
     if (KWayland::Client::XdgDecorationManager *manager = m_backend->display()->xdgDecorationManager()) {
         m_xdgDecoration.reset(manager->getToplevelDecoration(m_xdgShellSurface.get()));
         m_xdgDecoration->setMode(KWayland::Client::XdgDecoration::Mode::ServerSide);
@@ -152,31 +149,25 @@ WaylandOutput::WaylandOutput(const QString &name, WaylandBackend *backend)
         m_fractionalScale = wp_fractional_scale_manager_v1_get_fractional_scale(manager, *m_surface);
         wp_fractional_scale_v1_add_listener(m_fractionalScale, &s_fractionalScaleListener, this);
     }
-    m_viewport = wp_viewporter_get_viewport(backend->display()->viewporter(), *m_surface);
+    m_viewport = backend->display()->viewporter()->createViewport(*m_surface);
     setInformation(Information{
         .name = name,
         .model = name,
         .capabilities = caps,
     });
 
-    m_turnOffTimer.setSingleShot(true);
-    m_turnOffTimer.setInterval(dimAnimationTime());
-    connect(&m_turnOffTimer, &QTimer::timeout, this, [this] {
-        updateDpmsMode(DpmsMode::Off);
-    });
-
     m_configureThrottleTimer.setSingleShot(true);
     connect(&m_configureThrottleTimer, &QTimer::timeout, this, [this]() {
         applyConfigure(m_pendingConfigureSize, m_pendingConfigureSerial);
     });
 
-    connect(m_surface.get(), &KWayland::Client::Surface::frameRendered, this, [this]() {
-        Q_ASSERT(m_frame);
-        m_frame->presented(std::chrono::steady_clock::now().time_since_epoch(), PresentationMode::VSync);
-        m_frame.reset();
-    });
-
     updateWindowTitle();
+    if (auto toplevelIconManager = backend->display()->toplevelIconManager()) {
+        auto toplevelIcon = xdg_toplevel_icon_manager_v1_create_icon(toplevelIconManager);
+        xdg_toplevel_icon_v1_set_name(toplevelIcon, "kwin");
+        xdg_toplevel_icon_manager_v1_set_icon(toplevelIconManager, *m_xdgShellSurface, toplevelIcon);
+        xdg_toplevel_icon_v1_destroy(toplevelIcon);
+    }
 
     connect(m_xdgShellSurface.get(), &XdgShellSurface::configureRequested, this, &WaylandOutput::handleConfigure);
     connect(m_xdgShellSurface.get(), &XdgShellSurface::closeRequested, qApp, &QCoreApplication::quit);
@@ -186,11 +177,11 @@ WaylandOutput::WaylandOutput(const QString &name, WaylandBackend *backend)
 
 WaylandOutput::~WaylandOutput()
 {
-    if (m_presentationFeedback) {
-        wp_presentation_feedback_destroy(m_presentationFeedback);
-        m_presentationFeedback = nullptr;
+    m_frames.clear();
+    if (m_shortcutInhibition) {
+        zwp_keyboard_shortcuts_inhibitor_v1_destroy(m_shortcutInhibition);
     }
-    wp_viewport_destroy(m_viewport);
+    m_viewport.reset();
     m_xdgDecoration.reset();
     m_xdgShellSurface.reset();
     m_surface.reset();
@@ -251,6 +242,22 @@ static constexpr struct wp_presentation_feedback_listener s_presentationListener
     .discarded = handleDiscarded,
 };
 
+void WaylandOutput::handleFrame(void *data, wl_callback *callback, uint32_t time)
+{
+    auto output = reinterpret_cast<WaylandOutput *>(data);
+    auto it = std::ranges::find_if(output->m_frames, [callback](const auto &frame) {
+        return frame.frameCallback == callback;
+    });
+    if (it != output->m_frames.end()) {
+        // don't use the "time" argument for this, as it's in an unspecified base.
+        it->frameCallbackTime = std::chrono::steady_clock::now();
+    }
+}
+
+const wl_callback_listener WaylandOutput::s_frameCallbackListener{
+    .done = &WaylandOutput::handleFrame,
+};
+
 bool WaylandOutput::testPresentation(const std::shared_ptr<OutputFrame> &frame)
 {
     auto cursorLayers = Compositor::self()->backend()->compatibleOutputLayers(this) | std::views::filter([](OutputLayer *layer) {
@@ -262,6 +269,31 @@ bool WaylandOutput::testPresentation(const std::shared_ptr<OutputFrame> &frame)
     return true;
 }
 
+WaylandOutput::FrameData::FrameData(const std::shared_ptr<OutputFrame> &frame, struct wp_presentation_feedback *presentationFeedback, struct wl_callback *frameCallback)
+    : outputFrame(frame)
+    , presentationFeedback(presentationFeedback)
+    , frameCallback(frameCallback)
+{
+}
+
+WaylandOutput::FrameData::FrameData(FrameData &&move)
+    : outputFrame(std::move(move.outputFrame))
+    , presentationFeedback(std::exchange(move.presentationFeedback, nullptr))
+    , frameCallback(std::exchange(move.frameCallback, nullptr))
+    , frameCallbackTime(std::exchange(move.frameCallbackTime, std::nullopt))
+{
+}
+
+WaylandOutput::FrameData::~FrameData()
+{
+    if (presentationFeedback) {
+        wp_presentation_feedback_destroy(presentationFeedback);
+    }
+    if (frameCallback) {
+        wl_callback_destroy(frameCallback);
+    }
+}
+
 bool WaylandOutput::present(const QList<OutputLayer *> &layersToUpdate, const std::shared_ptr<OutputFrame> &frame)
 {
     auto cursorLayers = layersToUpdate | std::views::filter([](OutputLayer *layer) {
@@ -280,7 +312,7 @@ bool WaylandOutput::present(const QList<OutputLayer *> &layersToUpdate, const st
         m_surface->attachBuffer(buffer);
         m_mapped = true;
     }
-    wp_viewport_set_destination(m_viewport, geometry().width(), geometry().height());
+    m_viewport->setDestination(QSize(std::round(modeSize().width() / scale()), std::round(modeSize().height() / scale())));
     m_surface->setScale(1);
     // commit the subsurfaces before the main surface
     for (OutputLayer *layer : layersToUpdate) {
@@ -292,24 +324,21 @@ bool WaylandOutput::present(const QList<OutputLayer *> &layersToUpdate, const st
     if (m_backend->display()->tearingControl()) {
         m_renderLoop->setPresentationMode(frame->presentationMode());
     }
-    if (auto presentationTime = m_backend->display()->presentationTime()) {
-        m_presentationFeedback = wp_presentation_feedback(presentationTime, *m_surface);
-        wp_presentation_feedback_add_listener(m_presentationFeedback, &s_presentationListener, this);
-        m_surface->commit(KWayland::Client::Surface::CommitFlag::None);
-    } else {
-        m_surface->commit(KWayland::Client::Surface::CommitFlag::FrameCallback);
-    }
-    m_frame = frame;
+    FrameData frameData{
+        frame,
+        wp_presentation_feedback(m_backend->display()->presentationTime(), *m_surface),
+        wl_surface_frame(*m_surface),
+    };
+    wp_presentation_feedback_add_listener(frameData.presentationFeedback, &s_presentationListener, this);
+    wl_callback_add_listener(frameData.frameCallback, &s_frameCallbackListener, this);
+    m_surface->commit(KWayland::Client::Surface::CommitFlag::None);
+    m_frames.push_back(std::move(frameData));
     return true;
 }
 
 void WaylandOutput::frameDiscarded()
 {
-    m_frame.reset();
-    if (m_presentationFeedback) {
-        wp_presentation_feedback_destroy(m_presentationFeedback);
-        m_presentationFeedback = nullptr;
-    }
+    m_frames.pop_front();
 }
 
 void WaylandOutput::framePresented(std::chrono::nanoseconds timestamp, uint32_t refreshRate)
@@ -323,12 +352,17 @@ void WaylandOutput::framePresented(std::chrono::nanoseconds timestamp, uint32_t
         setState(next);
         m_renderLoop->setRefreshRate(m_refreshRate);
     }
-    m_frame->presented(timestamp, PresentationMode::VSync);
-    m_frame.reset();
-    if (m_presentationFeedback) {
-        wp_presentation_feedback_destroy(m_presentationFeedback);
-        m_presentationFeedback = nullptr;
+    const auto &frame = m_frames.front();
+    if (auto t = frame.frameCallbackTime) {
+        // NOTE that the frame callback gets signaled *after* the host compositor
+        // is done compositing the frame on the CPU side, not before!
+        // This is the best estimate we currently have for the commit deadline, but
+        // it should be replaced with something more accurate when possible.
+        const auto difference = timestamp - t->time_since_epoch();
+        m_renderLoop->setPresentationSafetyMargin(difference + std::chrono::milliseconds(1));
     }
+    frame.outputFrame->presented(timestamp, PresentationMode::VSync);
+    m_frames.pop_front();
 }
 
 void WaylandOutput::applyChanges(const OutputConfiguration &config)
@@ -348,6 +382,15 @@ void WaylandOutput::applyChanges(const OutputConfiguration &config)
     next.desiredModeRefreshRate = props->desiredModeRefreshRate.value_or(m_state.desiredModeRefreshRate);
     next.uuid = props->uuid.value_or(m_state.uuid);
     next.replicationSource = props->replicationSource.value_or(m_state.replicationSource);
+    next.dpmsMode = props->dpmsMode.value_or(m_state.dpmsMode);
+    if (next.dpmsMode != m_state.dpmsMode) {
+        if (next.dpmsMode == DpmsMode::On) {
+            m_renderLoop->uninhibit();
+        } else {
+            m_renderLoop->inhibit();
+        }
+    }
+    next.priority = props->priority.value_or(m_state.priority);
     setState(next);
 }
 
@@ -398,29 +441,6 @@ void WaylandOutput::init(const QSize &pixelSize, qreal scale, bool fullscreen)
     m_surface->commit(KWayland::Client::Surface::CommitFlag::None);
 }
 
-void WaylandOutput::setDpmsMode(DpmsMode mode)
-{
-    if (mode == DpmsMode::Off) {
-        if (!m_turnOffTimer.isActive()) {
-            Q_EMIT aboutToTurnOff(std::chrono::milliseconds(m_turnOffTimer.interval()));
-            m_turnOffTimer.start();
-        }
-    } else {
-        m_turnOffTimer.stop();
-        if (mode != dpmsMode()) {
-            updateDpmsMode(mode);
-            Q_EMIT wakeUp();
-        }
-    }
-}
-
-void WaylandOutput::updateDpmsMode(DpmsMode dpmsMode)
-{
-    State next = m_state;
-    next.dpmsMode = dpmsMode;
-    setState(next);
-}
-
 void WaylandOutput::handleConfigure(const QSize &size, XdgShellSurface::States states, quint32 serial)
 {
     if (!m_ready) {
@@ -428,7 +448,7 @@ void WaylandOutput::handleConfigure(const QSize &size, XdgShellSurface::States s
 
         applyConfigure(size, serial);
     } else {
-        // Output resizing is a resource intensive task, so the configure events are throttled.
+        // LogicalOutput resizing is a resource intensive task, so the configure events are throttled.
         m_pendingConfigureSerial = serial;
         m_pendingConfigureSize = size;
 
@@ -483,6 +503,7 @@ void WaylandOutput::lockPointer(Pointer *pointer, bool lock)
         m_pointerLock.reset();
         m_hasPointerLock = false;
         if (surfaceWasLocked) {
+            inhibitShortcuts(false);
             updateWindowTitle();
             Q_EMIT m_backend->pointerLockChanged(false);
         }
@@ -497,17 +518,38 @@ void WaylandOutput::lockPointer(Pointer *pointer, bool lock)
     }
     connect(m_pointerLock.get(), &LockedPointer::locked, this, [this]() {
         m_hasPointerLock = true;
+        inhibitShortcuts(true);
         updateWindowTitle();
         Q_EMIT m_backend->pointerLockChanged(true);
     });
     connect(m_pointerLock.get(), &LockedPointer::unlocked, this, [this]() {
         m_pointerLock.reset();
+        inhibitShortcuts(false);
         m_hasPointerLock = false;
         updateWindowTitle();
         Q_EMIT m_backend->pointerLockChanged(false);
     });
 }
 
+void WaylandOutput::inhibitShortcuts(bool inhibit)
+{
+    if (!inhibit) {
+        if (m_shortcutInhibition) {
+            zwp_keyboard_shortcuts_inhibitor_v1_destroy(m_shortcutInhibition);
+            m_shortcutInhibition = nullptr;
+        }
+        return;
+    }
+
+    auto *inhibitionManager = m_backend->display()->keyboardShortcutsInhibitManager();
+    if (!inhibitionManager) {
+        return;
+    }
+
+    Q_ASSERT(!m_shortcutInhibition);
+    m_shortcutInhibition = zwp_keyboard_shortcuts_inhibit_manager_v1_inhibit_shortcuts(inhibitionManager, *m_surface, *(m_backend->display()->seat()));
+}
+
 void WaylandOutput::setOutputLayers(std::vector<std::unique_ptr<OutputLayer>> &&layers)
 {
     m_layers = std::move(layers);
